{% extends "_base.html" %}
{% block title %}Success!{% endblock %}

{% block header %}
<link rel="stylesheet" href="/_matrix/static/client/register/style.css">
<script>
if (window.onAuthDone) {
    window.onAuthDone();
} else if (window.opener && window.opener.postMessage) {
    window.opener.postMessage("authDone", "*");
}
</script>
{% endblock %}

{% block body %}
<div>
    <p>Thank you</p>
    <p>You may now close this window and return to the application</p>
</div>

{% endblock %}
